/*
** Module   : BufferImpl.java
** Abstract : implementation of Buffer interface
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ----------------------Description------------------------
** 001 ECF 20130118 Created initial version. Implementation of the Buffer
**                  interface; provides services for buffer-related, Progress
**                  methods, attributes, and some language statements.
** 002 VMN 20130127 Added support for handle-based methods
**                  BUFFER-COPY (KW_BUF_COPY)
**                  BUFFER-COMPARE (KW_BUF_COMP)
**                  BUFFER-FIELD (KW_BUF_FLD)
**                  BUFFER-RELEASE (KW_BUF_REL)
**                  FIND-BY-ROWID (KW_FIND_BR)
**                  FIND-UNIQUE (KW_GET_UNI)
**                  FIND-FIRST (KW_FIND_1ST)
**                  FIND-LAST (KW_FIND_LST)
**
**                  and attributes
**
**                  ERROR (KW_ERROR)
**                  NUM-FIELDS (KW_NUM_FLD)
**                  NUM-ROWID (KW_ROWID)
**                  COLUMN-LABEL (KW_COL_LAB)
**                  LOCKED (KW_LOCKED)
**                  INITIAL (KW_INIT)
**                  AMBIGUOUS (KW_AMBIG)
*/

package com.goldencode.p2j.persist;

import com.goldencode.p2j.util.*;

/**
 * This class implements services representing Progress builtin functions,
 * methods, attributes, and language statements, which operate on record
 * buffers.
 * <p>
 * Converted business logic invokes these methods on DMO proxies via a
 * DMO-specific interface in the form <code>&lt;DMO interface&gt;.Buf</code>.
 * Each such interface extends both a custom interface of DMO-specific getter
 * and setter methods, as well as the {@link Buffer} interface, which defines
 * the API for these standard services.
 * <p>
 * Dynamic proxy classes for each DMO type extend this class, so that business
 * logic can invoke these standard services directly as method calls on DMO
 * proxy instances.
 * <p>
 * This class is intentionally package private, since it is not meant to be
 * extended outside of this package (and within this package only by dynamic
 * proxy class), nor instantiated directly.
 * <p>
 * This class implements {@link BufferReference} in order to grant its
 * instance methods access to the {@link RecordBuffer} object associated with
 * each instance of the DMO proxy class which extends this class. This access
 * is necessary for these methods to do their work. Even though {@link
 * #buffer()} is a public method, external packages will not be able to invoke
 * this method (except through reflection). This is by design, as business
 * logic is not meant to have access to <code>RecordBuffer</code> instances.
 * The default implementation of this method here throws an exception, as it
 * is never intended to be called, but it needs to be present in order for
 * this class to compile, since it is referenced by other methods. Each DMO
 * proxy class overrides this method to provide a working implementation.
 */
public abstract class BufferImpl
implements Buffer
{
   /**
    * Get the {@link RecordBuffer} instance associated with this object.
    * <p>
    * This implementation exists only to satisfy the Java compiler. This
    * method is only meaningful when a proxy DMO class is created by {@link
    * RecordBuffer}, which will implement this method to return an instance
    * of <code>RecordBuffer</code> associated with the proxy DMO instance.
    *
    * @return  Record buffer associated with this object.
    * 
    * @throws  UnsupportedOperationException
    *          always.
    */
   protected abstract RecordBuffer buffer();
   
   /**
    * Reports if this object is valid for use.  
    *
    * @return  <code>true</code> if we are valid (can be used).
    */
   public boolean isValid()
   {
      // TODO:  implement correctly
      return true;
   }
   
   /**
    * Reports if this object is unknown.  
    *
    * @return  <code>true</code> if object is unknown.
    */
   public boolean isUnknown()
   {
      // TODO:  implement correctly
      return false;
   }
   
  /**
   * Get the record ID of the current buffer's current record, if any.  P2J
   * implementation of the RECID function.
   * <p>
   * This method will return a 32-bit recid which represents the internal
   * row ID of the given record.  If the actual row ID (which is implemented
   * as a wider integer) cannot fit due to overflow, an error condition will
   * be raised.
   *
   * @return  Record ID of the underlying data record. If no record currently
   *          backs the buffer associated with <code>dmo</code>, a recid
   *          initialized as unknown value is returned.
   * 
   * @throws  ErrorConditionException
   *          if the actual record ID is too wide to fit within a 32-bit
   *          integer.
   */
  public recid recordID()
  {
     return buffer().recordID();
  }
  
  /**
   * Get the row ID of the current buffer's current record, if any.  P2J
   * implementation of the ROWID function.
   *
   * @return  Row ID of the underlying data record.  If no record currently
   *          backs the buffer associated with <code>dmo</code>, a
   *          <code>rowid</code> initialized as unknown value is returned.
   */
  public rowid rowID()
  {
     return buffer().rowID();
  }
  
   /**
    * Indicate whether a backing data record currently is available.
    *
    * @return  <code>true</code> if data is available, <code>false</code> if
    *          no record backs this buffer.
    */
   public logical available()
   {
      return (new logical(_available()));
   }
   
   /**
    * Indicate whether a backing data record currently is available.
    *
    * @return  <code>true</code> if data is available, <code>false</code> if
    *          no record backs this buffer.
    */
   public boolean _available()
   {
      return buffer().isAvailable();
   }
   
   /**
    * Checks if the record in the buffer has changed in database since the 
    * last <code>FIND CURRENT</code> or <code>GET CURRENT</code>.
    * P2J implementation of 4GL function <code>CURRENT-CHANGED</code>.
    *
    * @return  <code>true</code> if the database record differs from the
    *          value in memory buffer that was queried with the last
    *          <code>FIND</code> and <code>false</code> otherwise
    */
   public logical currentChanged()
   {
      return new logical(_currentChanged());
   }
   
   /**
    * Checks if the record in the buffer has changed in database since the 
    * last <code>FIND CURRENT</code> or <code>GET CURRENT</code>.
    * P2J implementation of 4GL function <code>CURRENT-CHANGED</code>.
    *
    * @return  <code>true</code> if the database record differs from the
    *          value in memory buffer that was queried with the last
    *          <code>FIND</code> and <code>false</code> otherwise
    */
   public boolean _currentChanged()
   {
      return buffer().isRecordChanged();
   }
   
   /**
    * Indicate whether the buffer associated currently contains a record, and
    * whether that record is transient.  A record is transient if it has been
    * newly created and has not yet been saved to the database.
    *
    * @return  <code>true</code> if a record exists and is new, else
    *          <code>false</code>.
    */
   public logical newlyCreated()
   {
      return (new logical(_newlyCreated()));
   }
   
   /**
    * Indicate whether the buffer associated currently contains a record, and
    * whether that record is transient.  A record is transient if it has been
    * newly created and has not yet been saved to the database.
    *
    * @return  <code>true</code> if a record exists and is new, else
    *          <code>false</code>.
    */
   public boolean _newlyCreated()
   {
      return buffer().isNewlyCreated();
   }
   
   /**
    * Indicate whether the last record retrieval attempt was denied because
    * the target record was locked.
    *
    * @return  <code>true</code> if the record was locked;  <code>false</code>
    *          if the attempt succeeded, or if this query has not yet
    *          attempted to retrieve a record.
    */
   public logical locked()
   {
      return (new logical(_locked()));
   }
   
   /**
    * Indicate whether the last record retrieval attempt was denied because
    * the target record was locked.
    *
    * @return  <code>true</code> if the record was locked;  <code>false</code>
    *          if the attempt succeeded, or if this query has not yet
    *          attempted to retrieve a record.
    */
   public boolean _locked()
   {
      return buffer().wasLocked();
   }
   
   /**
    * Indicate whether the last record retrieval attempt failed because the
    * query criteria was ambiguous.
    *
    * @return  <code>true</code> if the query was ambiguous;
    *          <code>false</code> if the attempt succeeded, or if no attempt
    *          has yet been made.
    */
   public logical ambiguous()
   {
      return (new logical(_ambiguous()));
   }
   
   /**
    * Indicate whether the last record retrieval attempt failed because the
    * query criteria was ambiguous.
    *
    * @return  <code>true</code> if the query was ambiguous;
    *          <code>false</code> if the attempt succeeded, or if no attempt
    *          has yet been made.
    */
   public boolean _ambiguous()
   {
      return buffer().wasAmbiguous();
   }
   
   /**
    * Create a new record in the buffer.
    *
    * @throws  ErrorConditionException
    *          if there is an error retrieving the next available primary key
    *          from the database or locking the record.
    */
   public void create()
   {
      buffer().create();
   }
   
   /**
    * Delete the record currently in the buffer, if any.
    *
    * @throws  ErrorConditionException
    *          if no record currently is loaded in the buffer, or if there is
    *          an error deleting the current record from the database.
    */
   public void delete()
   {
      try
      {
         buffer().delete();
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(exc);
      }
   }
   
   /**
    * Delete all records associated with this buffer, which match the criteria
    * specified by the given where clause.  Note that this where clause must
    * not contain DMO alias qualifiers.
    * <p>
    * <b>NOTE:</b>  the current implementation must not be invoked for a DMO
    * which is backed by a permanent table;  only temp-table bulk delete is
    * supported at this time.
    * 
    * @param   where
    *          An HQL where clause snippet which defines the restriction
    *          criteria to apply to the delete.  All references to properties
    *          in a DMO must be unqualified.
    * @param   args
    *          Query substitution parameters required by the where clause.
    * 
    * @throws  ErrorConditionException
    *          if there is an error deleting the records.
    * @throws  UnsupportedOperationException
    *          if invoked on a buffer associated with a permanent table.
    */
   public void delete(String where, Object ...args)
   {
      try
      {
         buffer().delete(where, args);
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(exc);
      }
   }
   
   /**
    * Bulk delete all records associated with the buffer.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the bulk delete operation.
    */
   public void deleteAll()
   {
      buffer().deleteAll();
   }
   
   /**
    * Validate and release the record currently in the buffer, if any.
    *
    * @throws  ErrorConditionException
    *          if the buffer fails validation.
    */
   public void release()
   {
      buffer().release();
   }
   
   /**
    * Validate the record currently in the buffer, if any.
    *
    * @throws  ErrorConditionException
    *          if the buffer fails validation.
    */
   public void validate()
   {
      buffer().validate();
   }
   
   /**
    * Open a read-only buffer scope (for trigger support).
    */
   public void openReadOnlyScope()
   {
      // TODO:  implement
   }

   /**
    * Conversion of BUFFER-FIELD() method (KW_BUF_FLD).
    *
    * Gets the handle to a particular buffer of a query or ProDataSet object.
    *
    * @param   fieldNumber
    *          field number. An INTEGER expression representing the sequence
    *          number of the field in the buffer.
    *
    * @return handle to a particular field in the buffer.
    */
   public BaseDataType bufferField(int fieldNumber)
   {
      // TODO:  implement
      return null;
   }

   /**
    * Conversion of BUFFER-FIELD() method (KW_BUF_FLD).
    *
    * Gets the handle to a particular buffer of a query or ProDataSet object.
    *
    * @param   fieldName
    *          field name. A CHARACTER string expression representing the name
    *          of the field in the buffer.
    *
    * @return handle to a particular field in the buffer.
    */
   public BaseDataType bufferField(String fieldName)
   {
      // TODO:  implement
      return null;
   }

   /**
    * Conversion of BUFFER-COPY() method (KW_BUF_COPY).
    *
    * This method copies any common fields, determined by name, data type,
    * and extent-matching, from the source buffer to the receiving buffer.
    * If there are fields in one buffer that do not exist in the other, they
    * are ignored. This method is used to accommodate temp-tables of joins.
    *
    * @return  <code>true</code> if BUFFER-COPY() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferCopy(handle bufHandle)
   {
      // TODO:  implement
      //RecordBuffer.copy((BufferImpl) bufHandle.get(), this, true);
      return new logical(true);
   }

   /**
    * Conversion of BUFFER-RELEASE() method (KW_BUF_REL).
    *
    * Releases a record from a buffer object. The BUFFER-RELEASE method
    * corresponds to the RELEASE statement.
    *
    * @return  <code>true</code> if BUFFER-RELEASE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferRelease()
   {
      // TODO:  implement
      //release();
      return new logical(true);
   }

   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record
    * into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the
    *          desired record.
    *
    * @return  <code>true</code> if it finds the record, else
    *          <code>false</code>.
    */
   public logical findByRowId(rowid id)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record
    * into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the
    *          desired record.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if it finds the record, else
    *          <code>false</code>.
    */
   public logical findByRowId(rowid id, int lock)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record
    * into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the
    *          desired record.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if it finds the record, else
    *          <code>false</code>.
    */
   public logical findByRowId(rowid id, integer lock)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record
    * into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the
    *          desired record.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if it finds the record, else
    *          <code>false</code>.
    */
   public logical findByRowId(rowid id, integer lock, integer nowait)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record
    * into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the
    *          desired record.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if it finds the record, else
    *          <code>false</code>.
    */
   public logical findByRowId(rowid id, int lock, integer nowait)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record
    * into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the
    *          desired record.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if it finds the record, else
    *          <code>false</code>.
    */
   public logical findByRowId(rowid id, int lock, int nowait)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record
    * into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the
    *          desired record.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if it finds the record, else
    *          <code>false</code>.
    */
   public logical findByRowId(rowid id, integer lock, int nowait)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of BUFFER-COMPARE() method (KW_BUF_COMP).
    *
    * This method does a rough compare of any common fields, determined by
    * name, data type, and extent-matching, between the source buffer and the
    * target buffer. The resulting logical value is either TRUE or FALSE as
    * a whole. A single field that does not compare causes the entire buffer
    * to return FALSE. If there are fields in one buffer that do not exist in
    * the other, they are ignored.
    *
    * @param   srcBufHandle
    *          An expression that evaluates to a buffer handle.
    *
    * @return  <code>true</code> if BUFFER-COMPARE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferCompare(handle srcBufHandle)
   {
      // TODO:  implement
      //logical result = new logical();
      //RecordBuffer.compare((BufferImpl) srcBufHandle.get(), this, result);
      return new logical(true);
   }

   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds,
    *          else <code>false</code>.
    */
   public logical findFirst(String whereClause)
   {
      // TODO:  implement
      //new FindQuery(this, "'" + whereClause + "'", null, null).first();
      return new logical(true);
   }

   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds,
    *          else <code>false</code>.
    */
   public logical findFirst(String whereClause, int lock)
   {
      // TODO:  implement
      //new FindQuery(this, "'" + whereClause + "'", null, null).first(lock);
      return new logical(true);
   }

   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds,
    *          else <code>false</code>.
    */
   public logical findFirst(String whereClause, integer lock)
   {
      // TODO:  implement
      //new FindQuery(this, "'" + whereClause + "'", null, null).first(lock);
      return new logical(true);
   }

   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds,
    *          else <code>false</code>.
    */
   public logical findFirst(String whereClause, int lock, int nowait)
   {
      // TODO:  implement
      //new FindQuery(this, "'" + whereClause + "'", null, null).first(lock, nowait);
      return new logical(true);
   }

   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds,
    *          else <code>false</code>.
    */
   public logical findFirst(String whereClause, integer lock, int nowait)
   {
      // TODO:  implement
      //new FindQuery(this, "'" + whereClause + "'", null, null).first(lock, nowait);
      return new logical(true);
   }

   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds,
    *          else <code>false</code>.
    */
   public logical findFirst(String whereClause, int lock, integer nowait)
   {
      // TODO:  implement
      //new FindQuery(this, "'" + whereClause + "'", null, null).first(lock, nowait);
      return new logical(true);
   }

   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds,
    *          else <code>false</code>.
    */
   public logical findFirst(String whereClause, integer lock, integer nowait)
   {
      // TODO:  implement
      //new FindQuery(this, "'" + whereClause + "'", null, null).first(lock, nowait);
      return new logical(true);
   }

   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    *
    * @return  <code>true</code> if FIND-LAST() succeeds,
    *          else <code>false</code>.
    */
   public logical findLast(String whereClause)
   {
      // TODO:  implement
      //new FindQuery(this, "'" + whereClause + "'", null, null).last();
      return new logical(true);
   }

   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if FIND-LAST() succeeds,
    *          else <code>false</code>.
    */
   public logical findLast(String whereClause, int lock)
   {
      // TODO:  implement
      //new FindQuery(this, "'" + whereClause + "'", null, null).last(lock);
      return new logical(true);
   }

   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if FIND-LAST() succeeds,
    *          else <code>false</code>.
    */
   public logical findLast(String whereClause, integer lock)
   {
      // TODO:  implement
      //new FindQuery(this, "'" + whereClause + "'", null, null).last(lock);
      return new logical(true);
   }

   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if FIND-LAST() succeeds,
    *          else <code>false</code>.
    */
   public logical findLast(String whereClause, int lock, int nowait)
   {
      // TODO:  implement
      //new FindQuery(this, "'" + whereClause + "'", null, null).last(lock, nowait);
      return new logical(true);
   }

   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if FIND-LAST() succeeds,
    *          else <code>false</code>.
    */
   public logical findLast(String whereClause, int lock, integer nowait)
   {
      // TODO:  implement
      //new FindQuery(this, "'" + whereClause + "'", null, null).last(lock, nowait);
      return new logical(true);
   }

   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if FIND-LAST() succeeds,
    *          else <code>false</code>.
    */
   public logical findLast(String whereClause, integer lock, int nowait)
   {
      // TODO:  implement
      //new FindQuery(this, "'" + whereClause + "'", null, null).last(lock, nowait);
      return new logical(true);
   }

   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if FIND-LAST() succeeds,
    *          else <code>false</code>.
    */
   public logical findLast(String whereClause, integer lock, integer nowait)
   {
      // TODO:  implement
      //new FindQuery(this, "'" + whereClause + "'", null, null).last(lock, nowait);
      return new logical(true);
   }

   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This method lets a user get a unique record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds,
    *          else <code>false</code>.
    */
   public logical findUnique(String whereClause)
   {
      // TODO:  implement
      //new FindQuery(this, "'" + whereClause + "'", null, null).unique();
      return new logical(true);
   }

   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This method lets a user get a unique record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds,
    *          else <code>false</code>.
    */
   public logical findUnique(String whereClause, int lock)
   {
      // TODO:  implement
      //new FindQuery(this, "'" + whereClause + "'", null, null).unique(lock);
      return new logical(true);
   }

   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This method lets a user get a unique record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds,
    *          else <code>false</code>.
    */
   public logical findUnique(String whereClause, integer lock)
   {
      // TODO:  implement
      //new FindQuery(this, "'" + whereClause + "'", null, null).unique(lock);
      return new logical(true);
   }

   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This method lets a user get a unique record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds,
    *          else <code>false</code>.
    */
   public logical findUnique(String whereClause, int lock, int nowait)
   {
      // TODO:  implement
      //new FindQuery(this, "'" + whereClause + "'", null, null).unique(lock, wait);
      return new logical(true);
   }

   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This method lets a user get a unique record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds,
    *          else <code>false</code>.
    */
   public logical findUnique(String whereClause, int lock, integer nowait)
   {
      // TODO:  implement
      //new FindQuery(this, "'" + whereClause + "'", null, null).unique(lock, wait);
      return new logical(true);
   }

   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This method lets a user get a unique record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds,
    *          else <code>false</code>.
    */
   public logical findUnique(String whereClause, integer lock, int nowait)
   {
      // TODO:  implement
      //new FindQuery(this, "'" + whereClause + "'", null, null).unique(lock, wait);
      return new logical(true);
   }

   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This method lets a user get a unique record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds,
    *          else <code>false</code>.
    */
   public logical findUnique(String whereClause, integer lock, integer nowait)
   {
      // TODO:  implement
      //new FindQuery(this, "'" + whereClause + "'", null, null).unique(lock, wait);
      return new logical(true);
   }

   /**
    * Conversion of ERROR attribute (KW_ERROR).
    *
    * Getter of ERROR attribute.
    *
    * @return  current value of ERROR attribute.
    */
   public logical error()
   {
      // TODO:  implement
      return null;
   }

   /**
    * Conversion of ERROR attribute (KW_ERROR).
    *
    * Getter of ERROR attribute
    *
    */
   public void changeError(logical error)
   {
      // TODO:  implement
   }

   /**
    * Conversion of QUERY-OFF-END attribute (KW_QRY_OFF).
    *
    * Getter of QUERY-OFF-END attribute.
    *
    * @return  current value of QUERY-OFF-END attribute.
    */
   public logical queryOfEnd()
   {
      // TODO:  implement
      return null;
   }

   /**
    * Conversion of NUM-FIELDS attribute (KW_NUM_FLD).
    *
    * Getter of NUM-FIELDS attribute.
    *
    * @return  current value of NUM-FIELDS attribute.
    */
   public integer numFields()
   {
      // TODO:  implement
      return null;
   }

   /**
    * Conversion of NUM-ROWID attribute (KW_ROWID).
    *
    * Getter of NUM-FIELDS attribute.
    *
    * @return  current value of NUM-FIELDS attribute.
    */
   public integer rowId()
   {
      // TODO:  implement
      return null;
   }

   /**
    * Conversion of LOCKED attribute (KW_LOCKED).
    *
    * Getter of LOCKED attribute.
    *
    * @return  current value of LOCKED attribute.
    */
   public character characterLocked()
   {
      // TODO:  implement
      return null;
   }

   /**
    * Conversion of INITIAL attribute (KW_INIT).
    *
    * Getter of INITIAL attribute.
    *
    * @return  current value of INITIAL attribute.
    */
   public character initial()
   {
      // TODO:  implement
      return null;
   }
}
